# Copyright 2023 Unai Lería Fortea

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
 
import __future__
from io import TextIOWrapper
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    import numpy as CNP

from .parameters import ParametersManager
from .util import *
import copy

class GenericModel:
    """Creates a compartimental model from a dictionary and setting an `evolve` method.
    """

    def get_all_params_names(self):
        """Returns a set of all parameters names, fixed or not.

        Returns:
            (set[str]): Set of all parameters names.
        """
        return set(list(self.param_to_index.keys()) + list(self.fixed_param_to_index.keys()))


    def __init__(self, configuration: dict[str, Any]):
        """Creates a model from the configuration given.

        Args:
            configuration (dict[str, Any]): Model configuration.
        """
        self.configuration: dict[str, Any] = copy.deepcopy(configuration)
        
        self.param_to_index: dict[str, int] = { k:i for i,k in enumerate(self.configuration["params"].keys()) }
        self.fixed_param_to_index: dict[str, int] = { k:i for i,k in enumerate(self.configuration["fixed_params"].keys()) }
        self.compartiment_name_to_index: dict[str, int] = { k:i for i,k in enumerate(self.configuration["compartiments"].keys()) }


    def populate_model_parameters(self, **kargs):
        """Populates params array. Assigns shortcuts to call them by their name as an attribute.
        """
        parameter_manager = ParametersManager(self.configuration, self)
        
        N_SIMULATIONS = self.configuration["simulation"]["n_simulations"]
        self.params = CNP.zeros(
            (len(self.configuration["params"]), N_SIMULATIONS), dtype=CNP.float64
        )
        self.fixed_params = CNP.zeros(
            (len(self.configuration["fixed_params"]), 1), dtype=CNP.float64
        )
        
        for param in self.configuration["params"].keys():
            setattr(self, param, self.params[self.param_to_index[param]])
            
        for fparam in self.configuration["fixed_params"].keys():
            setattr(self, fparam, self.fixed_params[self.fixed_param_to_index[fparam]])
            
        parameter_manager.populate_params(self.params, **kargs)
        parameter_manager.populate_fixed_params(self.fixed_params)
        
        
    def populate_model_compartiments(self, **kargs):
        """Populates compartiments array. Assigns shortcuts to call them by their name as an attribute.
        """
        N_SIMULATIONS = self.configuration["simulation"]["n_simulations"]
        self.state = CNP.zeros(
            (len(self.configuration["compartiments"]), N_SIMULATIONS), dtype=CNP.float64
        )
        self.log_diff = CNP.zeros((N_SIMULATIONS, 1), dtype=CNP.float64)
        
        for c,i in self.compartiment_name_to_index.items():
            C = self.configuration["compartiments"][c]
            initial_value = C["initial_value"]
            if isinstance(initial_value, str):
                if initial_value in self.param_to_index.keys():
                    self.state[i,:] = self.params[self.param_to_index[initial_value]]
                continue
            self.state[i,:] = initial_value
           
        for c,i in self.compartiment_name_to_index.items():
            C = self.configuration["compartiments"][c]
            minus = C.get("minus_compartiments", False)
            if not minus:
                continue
            if not isinstance(minus, list):
                minus = [minus]
            for m in minus:
                self.state[i,:] -= self.state[self.compartiment_name_to_index[m],:]
                
        for comp in self.configuration["compartiments"].keys():
            setattr(self, comp, self.state[self.compartiment_name_to_index[comp]])
                

    def evolve(self, step, *args, **kargs):
        """This method must be overwritten to complete the model initialization.

        Args:
            step (int): Step of simulation. Simulation ends when `step = simulation.n_steps`
        """
        ...


    def get_diff(self, step, reference, reference_mask):
        """Returns a value that represents the distance from the simulation to the reference.
        This function can be overwritten if other deffinition os distance is needed.
        
        Args:
            step (int): Step of simulation. Simulation ends when `step = simulation.n_steps`
            reference (list[list[float]]): Reference(s) values.
            reference_mask (list[int]): Mask to obtain simulation values to compare with the reference(s). 

        Returns:
            (list[float]): Distance from simulations to reference(s).
        """
        diff = CNP.absolute(CNP.take(self.state, reference_mask, 0).T-reference[step])
        return CNP.log(diff + 1)


    def _internal_run_(self, inner, inner_args: list, outer, outer_args:list,  reference, save_file:str, *args, **kargs):
        """Internal function that executes the model.

        Args:
            inner (function): Function to call in the main loop.
            inner_args (list): Args given to `inner`.
            outer (function): Function to call after the main loop.
            outer_args (list): Args given to `outer`.
            reference (list[list[float]]): Reference values used to compare with the simulation.
            save_file (str): Filename of path to file.
        """
        N_EXECUTIONS = self.configuration["simulation"]["n_executions"]
        
        for execution in range(N_EXECUTIONS):
            progress_bar(f"Model running: ", execution, N_EXECUTIONS, len=min(20, max(N_EXECUTIONS,5)))
            self.log_diff[:] = 0
            self.populate_model_parameters(**kargs)
            self.populate_model_compartiments(**kargs)
            
            for step in range(self.configuration["simulation"]["n_steps"]):
                inner(self, step, reference, *inner_args, **kargs)
            outer(self, *outer_args, execution_number=execution, **kargs)
            
        progress_bar(f"Model running: ", N_EXECUTIONS, N_EXECUTIONS, len=min(20, max(N_EXECUTIONS,5)), end='\n')


    def run_no_diff(self, save_file: str, *args, **kargs):
        """Runs the model without computing the diference any reference(s).

        Args:
            save_file (str): Filename of path to file.
        """
        self._internal_run_(
            self.evolve, args, 
            save_parameters_no_diff, (save_file, self.param_to_index.keys(),  self.params), 
            None, save_file, 
            *args, **kargs
        )


    def run(self, reference, save_file: str, *args, **kargs):
        """Runs the model computing the diference from the reference(s).

        Args:
            reference (list[list[float]]): Reference(s) values.
            save_file (str): Filename of path to file.
        """
        
        reference_mask = CNP.array([self.compartiment_name_to_index[c] for c in self.configuration["reference"]["compartiments"]])
        
        def inner(model, step, reference, reference_mask, *args, **kargs):
            model.evolve(model, step, *args, **kargs)
            self.log_diff[:] += model.get_diff(step, reference, reference_mask)
        
        def outer(model, save_file, *args, **kargs):
            best_params, best_log_diff = get_best_parameters(model.params, model.log_diff, model.configuration["results"]["save_percentage"])
            save_parameters(save_file, model.param_to_index.keys(), best_params, best_log_diff)
            
        self._internal_run_(
            inner, (reference_mask,), 
            outer, (save_file,), 
            reference, save_file, 
            *args, **kargs
        )



# Copyright 2023 Unai Lería Fortea

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
 
__version__ = "0.0.1"
_CUPY_MODE_: bool

from . import generic_model
from . import util
from . import parameters

def use_cupy():
    import cupy
    CNP = cupy
    _CUPY_MODE_= True
    update_packages(CNP)

def use_numpy():
    import numpy
    CNP = numpy
    _CUPY_MODE_= False
    update_packages(CNP)

def update_packages(CNP):
    generic_model.CNP = CNP
    util.CNP = CNP
    parameters.parameters_manager.CNP = CNP

use_numpy()

GenericModel = generic_model.GenericModel


# Copyright 2023 Unai Lería Fortea

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import __future__
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    import numpy as CNP

class ParametersManager:
    """Class used in model parameter population.
    """
    def __init__(self, configuration: dict[str, Any], model):
        self.configuration = configuration
        self.model = model
        
    def populate_params(self, params, **kargs):
        """Populates GenericModel.params array. If specific values are given in `kargs` those are used.

        Args:
            params (ndarray): Params array.

        Raises:
            Exception: If the model configuration could not fill all the array.
        """
        counter = 0
        for param_name, param_config in self.configuration["params"].items():
            param_index = self.model.param_to_index.get(param_name)
            if param_index is not None:
                counter += 1
                TYPE = param_config.get("type", "float64")

                if value := kargs.get(param_name, False):
                    params[param_index] = value
                    continue
                
                if TYPE == 'int':
                    params[param_index] = CNP.random.randint(
                        param_config["min"], 
                        param_config["max"]+1, 
                        self.configuration["simulation"]["n_simulations"])
                    
                if TYPE == 'float64':
                    params[param_index] = CNP.random.random(
                        self.configuration["simulation"]["n_simulations"]) \
                            * (param_config["max"] - param_config["min"]) \
                            + param_config["min"]

        if counter < len(self.model.param_to_index.keys()):
            raise Exception("Parameters array could not be correctly created with current options.")


    def populate_fixed_params(self, fixed_params, **kargs):
        """Populates GenericModel.fixed_params with the configuration values.
        If specific values are given in `kargs` those are used.

        Args:
            fixed_params (ndarray): Fixed parameters.

        Raises:
            Exception: If the model configuration could not fill all the array.
        """
        counter = 0
        for fparam_name, value in self.configuration["fixed_params"].items():
            fparam_index = self.model.fixed_param_to_index.get(fparam_name)
            if fparam_index is not None:
                counter += 1 
                fixed_params[fparam_index] = kargs.get(fparam_name, value)

        if counter < len(self.model.fixed_param_to_index.keys()):
            raise Exception("Fixed parameters array could not be correctly created with current options.")



# Copyright 2023 Unai Lería Fortea

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import __future__
from io import TextIOWrapper

from typing import TYPE_CHECKING

from math import ceil
import copy

if TYPE_CHECKING:
    import numpy as CNP


def get_best_parameters(params, log_diff, save_percentage):
    "Retuns the best `save_percentage`% `params` of the simulations given their `log_diff` with real data." 
    save_count: int = ceil(log_diff.size*save_percentage*0.01)
    saved_params = CNP.zeros((save_count, params.shape[0]), dtype=CNP.float64)
    saved_log_diff = CNP.zeros((save_count, 1), dtype=CNP.float64)

    log_diff_index_sorted = CNP.argpartition(log_diff, save_count, 0)[0:save_count]
    
    saved_params[:,:] = CNP.take(params, log_diff_index_sorted, 1).T
    saved_log_diff[:] = CNP.take(log_diff, log_diff_index_sorted)
    return saved_params, saved_log_diff


def progress_bar(prefix, progress, total, *, sufix="", end='\r', len=10):
    """Prints a progress bar on standar output.

    Args:
        prefix (str): Prefix to the progress bar.
        progress (int|float): Progress value.
        total (int|float): Total progess posible.
        sufix (str, optional): Sufix to the progess bar. Defaults to "".
        end (str, optional): End value, set to `\\n` at the end. Defaults to '\r'.
        len (int, optional): Length of progress bar. Defaults to 10.
    """
    per = len * progress/float(total)
    print(f"\r{prefix} -> ||{'▮'*int(per) + '▯'*(len-int(per))} ||{per*100/len:.2f}%  {sufix}", end=end)


def save_parameters_no_diff(file: str, params_names: list[str], params: list[list[float]], *, execution_number=0):
    """Saves the parameters with the given names without the diff column.

    Args:
        file (str): Filename or path to file.
        params_names (list[str]): Name of parameters.
        params (list[list[float]]): Parameters array.
        execution_number (int, optional): Number of the execution. If `0` the header is printed. Defaults to 0.
    """
    with open(file, 'a' if execution_number!=0 else 'w') as file_out:
        CNP.savetxt(file_out, params.T, delimiter=',', comments='', header=",".join(params_names) if execution_number==0 else "")


def save_parameters(file: str, params_names: list[str], params: list[list[float]], log_diff: list[float], *, execution_number=0):
    """Saves the parameters with the given names including the diff column.

    Args:
        file (str): Filename or path to file.
        params_names (list[str]): Name of parameters.
        params (list[list[float]]): Parameters array.
        log_diff (list[float]): Diff array.
        execution_number (int, optional): Number of the execution. If `0` the header is printed. Defaults to 0.
    """
    with open(file, 'a' if execution_number!=0 else 'w') as file_out:
        CNP.savetxt(file_out, CNP.concatenate((log_diff, params), 1) , delimiter=',', comments='', header=",".join(["log_diff", *params_names]) if execution_number==0 else "")

def load_parameters(file: str):
    """Loads parameters from file with the same format as `save_parameters` and `save_parameters_no_diff`.

    Args:
        file (str): Filename or path to file.

    Returns:
        (list[list[float]]): Parameters array. First index selects the column (parameter).
    """
    with open(file, 'r') as file_in:
        results = CNP.loadtxt(file_in, delimiter=',', skiprows=1).T
    return results


def get_model_sample_trajectory(model, *args, **kargs):
    """Executes the model with `n_simulations = 1` and `n_executions = 1`.
    Returns all the intermediate states and the parameters.

    Args:
        model (GenericModel): Model to execute.

    Returns:
        (list[list[float]], list[list[float]]): Tuple of all states history and corresponding params.
    """
    prev_config = copy.deepcopy(model.configuration)

    model.configuration["simulation"]["n_simulations"] = 1
    model.configuration["simulation"]["n_executions"] = 1
    
    model.populate_model_parameters(*args, **kargs)
    model.populate_model_compartiments(*args, **kargs)
    saved_state = CNP.zeros((model.configuration["simulation"]["n_steps"], model.state.shape[0]))
    for step in range(model.configuration["simulation"]["n_steps"]):
        model.evolve(model, step, *args, **kargs)
        saved_state[step] = model.state[:, 0]
        
    model.configuration.update(prev_config)
    return saved_state.T, model.params[:, 0]


def get_percentiles_from_results(model, results, p_minor=5, p_max=95, *args, **kargs):
    """Returns an array of percentils `p_minor=5`, median and `p_max=95` of the given model and results.

    Args:
        model (GenericModel): Model used to generate the `results`.
        results (list[list[float]]): Result parameters of `model` execution.
        p_minor (int, optional): Smaller percentile. Defaults to 5.
        p_max (int, optional): Bigger percentile. Defaults to 95.

    Returns:
        (list[int, int, list[float]]): First index represents the reference defined in `reference.compartiments`. \
            Second index represents  `p_minor`, median or `p_max=`. Final represents the step in the simulation.
    """
    reference_mask = CNP.array([model.compartiment_name_to_index[c] for c in model.configuration["reference"]["compartiments"]])
    
    results_no_diff = results[1:]
    results_percentiles = CNP.zeros((reference_mask.shape[0], 3, model.configuration["simulation"]["n_steps"]))
    
    prev_config = copy.deepcopy(model.configuration)

    model.configuration["simulation"]["n_simulations"] = results.shape[1]
    model.configuration["simulation"]["n_executions"] = 1
    
    model.populate_model_parameters(*args, **kargs)
    model.params[:] = results_no_diff[:]
    model.populate_model_compartiments(*args, **kargs)
    
    def inner(model, step, reference, reference_mask, *args, **kargs):
        model.evolve(model, step, *args, **kargs)
        aux = CNP.take(model.state, reference_mask, 0)
        
        results_percentiles[:, 0, step] += CNP.percentile(aux, p_minor, 1)
        results_percentiles[:, 1, step] += CNP.median(aux, 1)
        results_percentiles[:, 2, step] += CNP.percentile(aux, p_max, 1)
        
    def outer(model, *args, **kargs):
        ...
        
    model._internal_run_(
        inner, (reference_mask,), 
        outer, (), 
        None, None,
        *args, **kargs
    )
    model.configuration.update(prev_config)
    
    return results_percentiles
